<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Controller;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Entity\Controller\EntityListController;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\workspace_approval\WorkspaceApprovalPermissions;
use Drupal\workspace_approval\WorkspaceApprovalRequestReviewerListBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Builds workspace approval request listing pages.
 */
final class WorkspaceApprovalRequestListController extends EntityListController {

  public function __construct(
    private readonly WorkspaceApprovalPermissions $workspaceApprovalPermissions,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new self(
      $container->get('workspace_approval.permissions'),
    );
  }

  /**
   * Gets a listing page title.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The title.
   */
  public function reviewerListingTitle(): TranslatableMarkup {
    $builder = $this->getListBuilder();
    $build = $builder->render();
    return $build['table']['#title'] ?? $builder->getTitle();
  }

  /**
   * Builds a listing for the current user.
   */
  public function reviewerListing(): array {
    return $this->getListBuilder()->render();
  }

  /**
   * Gets the reviewer list builder and sets it up with the current user.
   */
  private function getListBuilder(): WorkspaceApprovalRequestReviewerListBuilder {
    $builder = $this->entityTypeManager()->getHandler('workspace_approval_request', 'list_builder_reviewer');
    \assert($builder instanceof WorkspaceApprovalRequestReviewerListBuilder);
    $builder->setReviewer($this->currentUser());
    return $builder;
  }

  /**
   * Checks access to the controller.
   *
   * Suitable for a _custom_access route callback.
   */
  public function access(AccountInterface $account): AccessResultInterface {
    foreach (\array_keys($this->workspaceApprovalPermissions->permissions()) as $permission) {
      if ($account->hasPermission($permission)) {
        return AccessResult::allowed();
      }
    }
    return AccessResult::neutral();
  }

}
