<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Entity;

use Drupal\Core\Entity\EditorialContentEntityBase;
use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityPublishedTrait;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\Plugin\DataType\BooleanData;
use Drupal\Core\TypedData\PrimitiveInterface;
use Drupal\user\Entity\User;
use Drupal\user\EntityOwnerTrait;
use Drupal\user\UserInterface;
use Drupal\workflows\Entity\Workflow;
use Drupal\workflows\TransitionInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workflows\WorkflowTypeInterface;
use Drupal\workspace_approval\Plugin\QueueWorker\RequesterNotificationQueueWorker;
use Drupal\workspace_approval\WorkspaceApprovalNotifications;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;
use Drupal\workspaces\Entity\Workspace;
use Drupal\workspaces\WorkspaceInterface;

/**
 * Defines the workspace approval request entity class.
 *
 * @ContentEntityType(
 *   id = "workspace_approval_request",
 *   label = @Translation("Workspace approval request"),
 *   label_collection = @Translation("Workspace approval requests"),
 *   label_singular = @Translation("workspace approval request"),
 *   label_plural = @Translation("workspace approval requests"),
 *   label_count = @PluralTranslation(
 *     singular = "@count workspace approval requests",
 *     plural = "@count workspace approval requests",
 *   ),
 *   bundle_label = @Translation("Workspace approval request type"),
 *   handlers = {
 *     "list_builder" = "Drupal\workspace_approval\WorkspaceApprovalRequestListBuilder",
 *     "list_builder_reviewer" = "Drupal\workspace_approval\WorkspaceApprovalRequestReviewerListBuilder",
 *     "views_data" = "Drupal\views\EntityViewsData",
 *     "form" = {
 *       "add" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestForm",
 *       "edit" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestForm",
 *       "approve" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestApproveForm",
 *       "reject" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestRejectForm",
 *       "dismiss" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestDismissForm",
 *       "delete" = "Drupal\Core\Entity\ContentEntityDeleteForm",
 *       "delete-multiple-confirm" = "Drupal\Core\Entity\Form\DeleteMultipleForm",
 *     },
 *     "route_provider" = {
 *       "html" = "Drupal\workspace_approval\Routing\WorkspaceApprovalRequestHtmlRouteProvider",
 *     },
 *     "access" = "Drupal\workspace_approval\WorkspaceApprovalRequestAccessControlHandler",
 *     "workspace" = "Drupal\workspaces\Entity\Handler\IgnoredWorkspaceHandler",
 *     "storage" = "Drupal\workspace_approval\WorkspaceApprovalRequestStorage",
 *   },
 *   base_table = "workspace_approval_request",
 *   data_table = "workspace_approval_request_field_data",
 *   revision_table = "workspace_approval_request_revision",
 *   revision_data_table = "workspace_approval_request_field_revision",
 *   admin_permission = "administer workspace_approval_request types",
 *   entity_keys = {
 *     "id" = "id",
 *     "revision" = "vid",
 *     "bundle" = "bundle",
 *     "uuid" = "uuid",
 *     "owner" = "uid",
 *     "published" = "active",
 *   },
 *   revision_metadata_keys = {
 *     "revision_user" = "revision_user",
 *     "revision_created" = "revision_created",
 *     "revision_log_message" = "revision_log_message",
 *   },
 *   show_revision_ui = false,
 *   token_type = "workspace_approval_request",
 *   links = {
 *     "add-form" = "/admin/config/workflow/workspaces/manage/{workspace}/workspace_approval/request/{workspace_approval_request_type}",
 *     "add-page" = "/admin/config/workflow/workspaces/manage/{workspace}/workspace_approval/request",
 *     "canonical" = "/admin/content/workspace_approval/{workspace_approval_request}",
 *     "dismiss-form" = "/admin/content/workspace_approval/{workspace_approval_request}/dismiss",
 *     "approve-form" = "/admin/content/workspace_approval/{workspace_approval_request}/approve",
 *     "reject-form" = "/admin/content/workspace_approval/{workspace_approval_request}/reject",
 *     "delete-form" = "/admin/content/workspace_approval/{workspace_approval_request}/delete",
 *   },
 *   bundle_entity_type = "workspace_approval_request_type",
 * )
 */
final class WorkspaceApprovalRequest extends EditorialContentEntityBase implements WorkspaceApprovalRequestInterface {

  use EntityChangedTrait;
  use EntityOwnerTrait;
  use EntityPublishedTrait;

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage): void {
    parent::preSave($storage);
    if (\is_null($this->getOwnerId())) {
      // If no owner has been set explicitly, make the anonymous user the owner.
      $this->setOwnerId(0);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function label(): TranslatableMarkup {
    $bundle_info = $this->entityTypeBundleInfo()->getBundleInfo($this->getEntityTypeId());
    $bundle_label = $bundle_info[$this->bundle()]['label'];
    $workspace = $this->getSubject()->label();
    return new TranslatableMarkup('@workspace workspace @bundle_label', [
      '@workspace' => $workspace,
      '@bundle_label' => \strtolower((string) $bundle_label),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getReviewer(): UserInterface {
    $reviewer = $this->get('reviewer')->entity;
    // @todo remove the conditional block below before a stable release. See https://www.drupal.org/project/workspace_approval/issues/3483162.
    if (!$reviewer instanceof UserInterface) {
      $warning = \sprintf('An orphaned workspace approval request was detected (ID: %d). Have you run database updates?', $this->id());
      \Drupal::logger('workspace_approval')->warning($warning);
      $anonymous = User::load(0);
      \assert($anonymous instanceof UserInterface);
      return $anonymous;
    }
    return $reviewer;
  }

  /**
   * {@inheritdoc}
   */
  public function getReviewerId(): int {
    $reviewer = $this->get('reviewer')->first();
    \assert($reviewer instanceof EntityReferenceItem);
    $targetId = $reviewer->get('target_id');
    \assert($targetId instanceof PrimitiveInterface);
    return $targetId->getCastedValue();
  }

  /**
   * {@inheritdoc}
   */
  public function getSubject(): WorkspaceInterface {
    $workspace = $this->get('subject')->entity;
    // @todo remove the conditional block below before a stable release. See https://www.drupal.org/project/workspace_approval/issues/3483162.
    if (!$workspace instanceof WorkspaceInterface) {
      $warning = \sprintf('An orphaned workspace approval request was detected (ID: %d). Have you run database updates?', $this->id());
      \Drupal::logger('workspace_approval')->warning($warning);
      return Workspace::create([
        'id' => '00000000-0000-0000-0000-000000000000',
        'uuid' => '00000000-0000-0000-0000-000000000000',
        'label' => (string) new TranslatableMarkup('Removed'),
        'uid' => 0,
      ]);
    }
    return $workspace;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus(): ?bool {
    $status = $this->get('approved');
    \assert($status instanceof FieldItemListInterface);
    if ($status->isEmpty()) {
      return NULL;
    }
    $approved = $status->first()?->get('value');
    \assert($approved instanceof BooleanData);
    return $approved->getCastedValue();
  }

  /**
   * {@inheritdoc}
   */
  public function isApproved(): bool {
    return $this->getStatus() === TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function isRejected(): bool {
    return $this->getStatus() === FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function isPending(): bool {
    return $this->getStatus() === NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getMessageTemplate(WorkspaceApprovalNotifications $type): ?string {
    return $this->getApprovalType()->messageTemplate($type);
  }

  /**
   * {@inheritdoc}
   */
  public function getApprovalType(): WorkspaceApprovalRequestType {
    $request_type = $this->entityTypeManager()->getStorage('workspace_approval_request_type')->load($this->bundle());
    \assert($request_type instanceof WorkspaceApprovalRequestType);
    return $request_type;
  }

  /**
   * {@inheritdoc}
   */
  public function isApprovalRequired(?WorkflowInterface $workflow = NULL, ?TransitionInterface $transition = NULL): bool {
    if ($this->isApproved()) {
      // An approved request will never block a transition.
      return FALSE;
    }
    $workspace = $this->getSubject();
    $approvalType = $this->getApprovalType();
    $guardedWorkflowId = $approvalType->workflowId();
    $guardedTransitionIds = $approvalType->transitionIds();
    \assert(\is_string($guardedWorkflowId));
    if (empty($workflow)) {
      // Check the configured workflow if no workflow was given as an argument.
      $workflow = Workflow::load($guardedWorkflowId);
      \assert($workflow instanceof WorkflowInterface);
    }
    elseif ($workflow->id() !== $guardedWorkflowId) {
      // A workflow was given, but it doesn't match the guarded workflow.
      return FALSE;
    }
    $workflowPlugin = $workflow->getTypePlugin();
    \assert($workflowPlugin instanceof WorkflowTypeInterface);
    if (!empty($transition)) {
      // Ensure that the transition given belongs to the workflow to check.
      if (!$workflowPlugin->hasTransition($transition->id())) {
        throw new \InvalidArgumentException(\sprintf('The %s workflow does not have a %s transition', $guardedWorkflowId, $transition->id()));
      }
      if (!\in_array($transition->id(), $guardedTransitionIds, TRUE)) {
        // The transition is not guarded by this approval type.
        return FALSE;
      }
    }
    // If no transition was given as an argument, then check the transitions
    // guarded by this approval type.
    $transitions = \array_map(
      static fn (TransitionInterface $transition) => $transition->id(),
      \array_map($workflowPlugin->getTransition(...), $guardedTransitionIds),
    );
    // At this point, the workflow and transition to check have been validated
    // an the only remaining check is whether the workspace has the guarded
    // transition available in its current state.
    $workspaceState = \entity_workflow_get_entity_state($workspace, $guardedWorkflowId);
    \assert(!empty($workspaceState));
    $availableTransitions = $workflowPlugin->getTransitionsForState($workspaceState);
    // This approval type is only required when its guarded transition is
    // currently available.
    return !empty(\array_intersect($transitions, \array_keys($availableTransitions)));
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type): array {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $fields */
    $fields = parent::baseFieldDefinitions($entity_type);

    $fields[$entity_type->getRevisionMetadataKey('revision_log_message')]
      ->setLabel(new TranslatableMarkup('Remarks'))
      ->setDescription(new TranslatableMarkup('Provide a brief explanation of your review.'))
      ->setDisplayConfigurable('view', FALSE)
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'string',
        'weight' => 100,
      ]);

    $fields += self::ownerBaseFieldDefinitions($entity_type);
    $fields[$entity_type->getKey('owner')]
      ->setLabel(new TranslatableMarkup('Requester'))
      ->setRequired(TRUE)
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'author',
        'weight' => 70,
      ]);

    $fields += self::publishedBaseFieldDefinitions($entity_type);
    $fields[$entity_type->getKey('published')]
      ->setLabel(new TranslatableMarkup('Status'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE)
      ->setRevisionable(TRUE)
      ->setDefaultValue(TRUE)
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'boolean',
        'settings' => ['format' => 'default'],
        'weight' => 0,
      ])
      ->setSetting('on_label', 'Active')
      ->setSetting('off_label', 'Dismissed');

    $fields['approved'] = BaseFieldDefinition::create('boolean')
      ->setLabel(new TranslatableMarkup('Approval status'))
      ->setDisplayConfigurable('form', FALSE)
      ->setRevisionable(TRUE)
      ->setSetting('on_label', 'Approved')
      ->setSetting('off_label', 'Rejected')
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'boolean',
        'settings' => ['format' => 'default'],
        'weight' => 0,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['subject'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(new TranslatableMarkup('Review of'))
      ->setRequired(TRUE)
      ->setSetting('target_type', 'workspace')
      ->setDefaultValueCallback(self::class . '::getDefaultSubject')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'entity_reference_label',
        'weight' => 0,
      ])
      ->setDisplayConfigurable('view', TRUE);

    $fields['reviewer'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(new TranslatableMarkup('Reviewer'))
      ->setRequired(TRUE)
      ->setDescription('The user who will approve or reject the content')
      ->setSetting('target_type', 'user')
      ->setSetting('handler', 'workspace_approval:reviewer')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('form', [
        'type' => 'select2_entity_reference',
      ])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'author',
        'weight' => 90,
      ]);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(new TranslatableMarkup('Requested on'))
      ->setDescription(new TranslatableMarkup('The time that the workspace approval request was created.'))
      ->setDisplayOptions('view', [
        'label' => 'inline',
        'type' => 'timestamp',
        'weight' => 80,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(new TranslatableMarkup('Updated'))
      ->setDescription(new TranslatableMarkup('The time that the workspace approval request was last edited.'));

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function postSave(EntityStorageInterface $storage, $update = TRUE): void {
    if ($update) {
      \Drupal::queue(RequesterNotificationQueueWorker::ID)->createItem([
        'workspace_approval_request_id' => $this->id(),
      ]);
    }
    $this->invalidateTagsOnSave($update);
  }

  /**
   * {@inheritdoc}
   */
  protected function urlRouteParameters($rel): array {
    $parameters = parent::urlRouteParameters($rel);
    return $parameters + [
      'workspace' => $this->getSubject()->id(),
    ];
  }

  /**
   * Assigns this approval request to a particular workspace.
   *
   * @return string|null
   *   The workspace ID or NULL if a workspace cannot be determined.
   */
  public static function getDefaultSubject(): ?string {
    return \Drupal::routeMatch()->getRawParameter('workspace');
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTagsToInvalidate(): array {
    // Add the subject entity's tags so that when a request is approved or
    // rejected, the entity's workflow transition's local actions will also be
    // invalidated. This ensures that the available transition buttons are kept
    // up to date as requests are updated.
    return [
      ...parent::getCacheTagsToInvalidate(),
      ...$this->getSubject()->getCacheTagsToInvalidate(),
    ];
  }

}
