<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Entity;

use Drupal\Core\Config\Entity\ConfigEntityBundleBase;
use Drupal\workspace_approval\WorkspaceApprovalNotifications;

/**
 * Defines the workspace approval request type configuration entity.
 *
 * @ConfigEntityType(
 *   id = "workspace_approval_request_type",
 *   label = @Translation("Workspace approval request type"),
 *   label_collection = @Translation("Workspace approval request types"),
 *   label_singular = @Translation("workspace approval request type"),
 *   label_plural = @Translation("workspace approval requests types"),
 *   label_count = @PluralTranslation(
 *     singular = "@count workspace approval requests type",
 *     plural = "@count workspace approval requests types",
 *   ),
 *   handlers = {
 *     "form" = {
 *       "add" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestTypeForm",
 *       "edit" = "Drupal\workspace_approval\Form\WorkspaceApprovalRequestTypeForm",
 *       "delete" = "Drupal\Core\Entity\EntityDeleteForm",
 *     },
 *     "list_builder" = "Drupal\workspace_approval\WorkspaceApprovalRequestTypeListBuilder",
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\AdminHtmlRouteProvider",
 *     },
 *   },
 *   admin_permission = "administer workspace approval request types",
 *   bundle_of = "workspace_approval_request",
 *   config_prefix = "workspace_approval_request_type",
 *   token_type = "workspace_approval_request_type",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "workflow" = "workflow",
 *     "guarded_transitions" = "guarded_transitions",
 *     "message_templates" = "message_templates",
 *     "uuid" = "uuid",
 *   },
 *   links = {
 *     "add-form" = "/admin/config/workflow/approval_request_types/add",
 *     "edit-form" = "/admin/config/workflow/approval_request_types/manage/{workspace_approval_request_type}",
 *     "delete-form" = "/admin/config/workflow/approval_request_types/manage/{workspace_approval_request_type}/delete",
 *     "collection" = "/admin/config/workflow/approval_request_types",
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "workflow",
 *     "guarded_transitions",
 *     "message_templates",
 *     "uuid",
 *   },
 * )
 */
final class WorkspaceApprovalRequestType extends ConfigEntityBundleBase {

  /**
   * The machine name of this workspace approval request type.
   */
  protected string $id;

  /**
   * The human-readable name of the workspace approval request type.
   */
  protected string $label;

  /**
   * The workflow for this review type.
   */
  protected string $workflow;

  /**
   * The transitions to which this review type applies.
   */
  protected array $guarded_transitions;

  /**
   * The message templates to use for specific notification types.
   */
  protected ?array $message_templates;

  /**
   * The workflow ID to which this review type applies.
   *
   * @return string|null
   *   The workflow ID or NULL if this is a new entity and the workflow has not
   *   yet been selected.
   */
  public function workflowId(): ?string {
    return $this->workflow ?? NULL;
  }

  /**
   * The workflow transitions to which this review type applies.
   *
   * @return string[]
   *   An array of transition machine names.
   */
  public function transitionIds(): array {
    return $this->guarded_transitions ?? [];
  }

  /**
   * The workflow transition to which this review type applies.
   *
   * @return string|null
   *   The message template to use for the given notification message type or
   *   NULL if no message should be sent.
   */
  public function messageTemplate(WorkspaceApprovalNotifications $eventType): ?string {
    return $this->message_templates[$eventType->value] ?? NULL;
  }

}
