<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Form;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityReferenceSelection\SelectionInterface;
use Drupal\Core\Entity\EntityReferenceSelection\SelectionPluginManagerInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequest;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Drupal\workspace_approval\Plugin\EntityReferenceSelection\ReviewerSelection;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;
use Drupal\workspaces\WorkspaceInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Helper class for Drupal form alter hook implementations.
 */
final class EntityWorkflowFormAlters implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * Constructs a new EntityWorkflowFormAlters object.
   */
  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly RouteMatchInterface $routeMatch,
    private readonly SelectionPluginManagerInterface $selectionPluginManager,
    TranslationInterface $stringTranslation,
  ) {
    $this->stringTranslation = $stringTranslation;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new self(
      $container->get('entity_type.manager'),
      $container->get('current_route_match'),
      $container->get('plugin.manager.entity_reference_selection'),
      $container->get('string_translation'),
    );
  }

  /**
   * Type-hinted helper to wrap \Drupal::classResolver.
   */
  public static function getInstance(): self {
    $instance = \Drupal::classResolver(self::class);
    \assert($instance instanceof self);
    return $instance;
  }

  /**
   * Implements hook_entity_workflow_form.
   */
  public function entityWorkflowForm(array &$form, FormStateInterface $form_state): void {
    // Get the transition for the current form.
    $workflow = $this->routeMatch->getParameter('workflow');
    \assert($workflow instanceof WorkflowInterface);
    // Get the workspace that is will be the subject of any assigned reviews.
    $reviewSubject = $this->routeMatch->getParameter($this->routeMatch->getParameter('entity_type_id'));
    \assert($reviewSubject instanceof WorkspaceInterface);
    // Get the transitions that are available on the form.
    $transitions = \array_keys($form['workflow']['transition']['#options'] ?? []);
    $requestTypesByTransition = [];
    foreach ($transitions as $transition) {
      // Each transition may lead to a different state and those states may
      // require different types of reviews to "leave".
      $requestTypesByTransition[$transition] = $this->getApprovalRequestTypes($workflow, (string) $transition);
    }
    // Add an approver section to the form.
    $form['workflow']['#title'] = 'Assign reviews';
    // Each workspace_approval_request entity gets a form element.
    foreach ($requestTypesByTransition as $transition => $requestTypes) {
      $form['workflow'][$transition . '__reviewers'] = $this->buildReviewerSelects($reviewSubject, $requestTypes);
      foreach ($form['workflow'][$transition . '__reviewers'] as &$select) {
        $select['#states']['visible'][':input[name="transition"]'] = ['value' => $transition];
        $select['#states']['required'][':input[name="transition"]'] = ['value' => $transition];
        $select['#visible'] = FALSE;
        $select['#element_validate'][] = [$this, 'validateEntityWorkflowForm'];
      }
    }
    $form['workflow']['actions']['submit']['#validate'][] = [$this, 'validateEntityWorkflowForm'];
    $form['workflow']['actions']['submit']['#submit'][] = [$this, 'submitEntityWorkflowForm'];
    $form['workflow']['actions']['submit']['#submit'][] = [$form_state->getFormObject(), 'submitForm'];
  }

  /**
   * Validation handler for the entity_workflow_form.
   */
  public function validateEntityWorkflowForm(array &$element, FormStateInterface $form_state): void {
    $workflow = $this->routeMatch->getParameter('workflow');
    \assert($workflow instanceof WorkflowInterface);
    $transition = $form_state->getValue('transition');
    if (!empty($element['workflow'][$transition . '__reviewers'])) {
      $requestTypes = \array_keys($this->getApprovalRequestTypes($workflow, (string) $transition));
      foreach ($requestTypes as $requestType) {
        $selections = \array_filter(
          $form_state->getValue($requestType),
          static fn ($id) => $id !== 0,
        );
        if (empty($selections)) {
          $form_state->setError($element['workflow'][$transition . '__reviewers'][$requestType], new TranslatableMarkup('%name is required.', ['%name' => $element['#title']]));
        }
      }
    }
  }

  /**
   * Submit handler for the entity_workflow_form.
   */
  public function submitEntityWorkflowForm(array &$form, FormStateInterface $form_state): void {
    if (!$form_state->isValidationComplete()) {
      return;
    }

    $transition_id = $form_state->getValue('transition');
    if ($transition_id === 'back_to_draft') {
      $entityStorage = $this->entityTypeManager->getStorage('workspace_approval_request');
      // Clear out all the approvals by marking them as unpublished.
      $approvals = $entityStorage->loadMultiple($entityStorage->getQuery()->accessCheck()->execute());
      foreach ($approvals as $approvalRequest) {
        \assert($approvalRequest instanceof WorkspaceApprovalRequestInterface);
        $approvalRequest->setUnpublished()->save();
      }
      return;
    }

    // Get the transition for the current form.
    $workflow = $this->routeMatch->getParameter('workflow');
    \assert($workflow instanceof WorkflowInterface);

    foreach ($this->getApprovalRequestTypes($workflow, $transition_id) as $requestType) {
      $uids = \array_keys(\array_filter($form_state->getValue((string) $requestType->id())));
      foreach ($uids as $uid) {
        $request = WorkspaceApprovalRequest::create([
          'bundle' => $requestType->id(),
          'reviewer' => $uid,
        ]);
        $request->setPublished()->save();
      }
    }
  }

  /**
   * Implements hook_entity_workflow_simple_transition_form.
   */
  public function entityWorkflowSimpleTransitionForm(array &$form, FormStateInterface $form_state): void {
    // Get the transition for the current form.
    $workflow = $this->routeMatch->getParameter('workflow');
    \assert($workflow instanceof WorkflowInterface);
    $transition_id = $this->routeMatch->getParameter('transition_id');
    $form['actions']['submit']['#submit'][] = [$this, 'submitWorkflowSimpleTransitionForm'];
    $form['actions']['submit']['#submit'][] = [$form_state->getFormObject(), 'submitForm'];
    $requestTypes = $this->getApprovalRequestTypes($workflow, $transition_id);
    if (empty($requestTypes)) {
      return;
    }
    // Add an approver section to the form.
    $form['approvers']['#type'] = 'fieldset';
    $form['approvers']['#title'] = 'Assign reviews';
    $form['approvers']['#open'] = TRUE;
    $reviewSubject = $this->routeMatch->getParameter($this->routeMatch->getParameter('entity_type_id'));
    // Each workspace_approval_request entity gets a form element.
    $form['approvers'] += $this->buildReviewerSelects($reviewSubject, $requestTypes);
  }

  /**
   * Builds a select list options array.
   */
  private function buildReviewerSelects(WorkspaceInterface $reviewSubject, array $requestTypes): array {
    $elements = [];
    // Each workspace_approval_request entity gets a form element.
    foreach ($requestTypes as $reviewType) {
      $selection = $this->selectionPluginManager->createInstance(
        'workspace_approval:reviewer',
        [
          ReviewerSelection::INCLUDE_ACTIVE_REVIEWERS_KEY => TRUE,
          'request_type' => $reviewType->id(),
          'subject' => $reviewSubject,
        ],
      );
      \assert($selection instanceof SelectionInterface);
      $candidates = $selection->getReferenceableEntities()['user'] ?? [];
      $active = $this->getActiveReviewerIds($reviewSubject, $reviewType);
      $elements[$reviewType->id()] = [
        '#type' => 'checkboxes',
        '#title' => $reviewType->label(),
        '#options' => $candidates,
        '#default_value' => $active,
      ];
      foreach ($active as $pendingId) {
        $elements[$reviewType->id()][$pendingId]['#disabled'] = TRUE;
      }
      $args = ['%reviewType' => $reviewType->label()];
      $elements[$reviewType->id()]['#description'] = !empty($candidates)
        ? $this->t('Select users to approve a %reviewType.', $args)
        : $this->t('There are no users with permission to approve %reviewType requests.', $args);
    }
    return $elements;
  }

  /**
   * Gets all user IDs that have an active request for the configured subject.
   *
   * @return array
   *   An array of user IDs identifying users that have an active request for
   *   the configured review type and subject.
   */
  private function getActiveReviewerIds(WorkspaceInterface $reviewSubject, WorkspaceApprovalRequestType $requestType): array {
    return \array_map(
      static fn (WorkspaceApprovalRequest $request) => $request->getReviewerId(),
      $this->getActiveRequests($reviewSubject, $requestType),
    );
  }

  /**
   * Gets all user IDs that have an active request for the configured subject.
   *
   * @return array
   *   An array of user IDs identifying users that have an active request for
   *   the configured review type and subject.
   */
  private function getActiveRequests(WorkspaceInterface $reviewSubject, WorkspaceApprovalRequestType $requestType): array {
    $requestStorage = $this->entityTypeManager->getStorage('workspace_approval_request');
    $entityType = $this->entityTypeManager->getDefinition('workspace_approval_request');
    \assert($entityType instanceof EntityTypeInterface);
    $bundleKey = $entityType->getKey('bundle');
    $activeKey = $entityType->getKey('published');
    \assert(\is_string($bundleKey) && \is_string($activeKey));
    return $requestStorage->loadMultiple($requestStorage->getQuery()
      ->condition('subject', $reviewSubject->id())
      ->condition($bundleKey, $requestType->id())
      ->condition($activeKey, TRUE)
      ->accessCheck(FALSE)
      ->execute());
  }

  /**
   * Submit handler for the entity_workflow_simple_transition_form.
   */
  public function submitWorkflowSimpleTransitionForm(array &$form, FormStateInterface $form_state): void {
    if (!$form_state->isValidationComplete()) {
      return;
    }
    $transition_id = $this->routeMatch->getParameter('transition_id');
    if ($transition_id === 'back_to_draft') {
      $entityStorage = $this->entityTypeManager->getStorage('workspace_approval_request');
      // Clear out all the approvals by marking them as unpublished.
      $approvals = $entityStorage->loadMultiple($entityStorage->getQuery()->accessCheck()->execute());
      foreach ($approvals as $approvalRequest) {
        \assert($approvalRequest instanceof WorkspaceApprovalRequestInterface);
        $approvalRequest->setUnpublished()->save();
      }
      return;
    }
    $reviewSubject = $this->routeMatch->getParameter($this->routeMatch->getParameter('entity_type_id'));
    // Get the transition for the current form.
    $workflow = $this->routeMatch->getParameter('workflow');
    \assert($workflow instanceof WorkflowInterface);
    foreach ($this->getApprovalRequestTypes($workflow, $transition_id) as $requestType) {
      $uids = \array_keys(\array_filter($form_state->getValue((string) $requestType->id())));
      $pending_requests = $this->getActiveRequests($reviewSubject, $requestType);
      $pending_requests_by_uid = \array_combine(\array_map(static fn (WorkspaceApprovalRequestInterface $request) => $request->getReviewerId(), $pending_requests), $pending_requests);
      foreach ($uids as $uid) {
        if (\array_key_exists($uid, $pending_requests_by_uid)) {
          $request = $pending_requests_by_uid[$uid];
        }
        else {
          $request = WorkspaceApprovalRequest::create([
            'bundle' => $requestType->id(),
            'reviewer' => $uid,
          ]);
        }
        $request->setPublished()->save();
      }
    }
  }

  /**
   * Loads all workspace approval request types for the given transition.
   *
   * @param \Drupal\workflows\WorkflowInterface $workflow
   *   The workflow in which the transition is found.
   * @param string $transition_id
   *   The transition for which request types should be loaded.
   *
   * @return \Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType[]
   *   An array of workspace approval request type objects.
   */
  private function getApprovalRequestTypes(WorkflowInterface $workflow, string $transition_id): array {
    $workflow_type = $workflow->getTypePlugin();
    $transition = $workflow_type->getTransition($transition_id);
    // Figure out which state the transition targets and which transitions are
    // available from that state.
    $target_state = $transition->to()->id();
    $next_transitions = $workflow_type->getTransitionsForState($target_state);
    // Find all the workspace approval request types that guard the transitions
    // that will be available in the next state. This allows the user to pick
    // reviewers to approve the subsequent transition.
    $storage = $this->entityTypeManager->getStorage('workspace_approval_request_type');
    \assert($storage instanceof EntityStorageInterface);
    $query = $storage
      ->getQuery()
      ->condition('workflow', $workflow->id())
      ->condition('guarded_transitions.*', \array_keys($next_transitions), 'IN')
      ->accessCheck(FALSE);
    /** @var \Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType[] $requestTypes */
    $requestTypes = $storage->loadMultiple($query->execute());
    return $requestTypes;
  }

}
