<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Form;

use Drupal\Component\Assertion\Inspector;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Entity\ContentEntityConfirmFormBase;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\entity_workflow\Entity\WorkflowTransitionLog;
use Drupal\entity_workflow\EntityWorkflowTransition;
use Drupal\workflows\Entity\Workflow;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;

/**
 * Base class for forms confirming an operation on a workflow approval request.
 */
abstract class WorkspaceApprovalRequestConfirmForm extends ContentEntityConfirmFormBase {

  /**
   * The guarded workflow transition.
   */
  protected ?WorkflowInterface $workflow = NULL;

  /**
   * The guarded workflow transitions.
   */
  protected array $transitions = [];

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityRepositoryInterface $entity_repository, EntityTypeBundleInfoInterface $entity_type_bundle_info, TimeInterface $time) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    \assert(!empty($this->operation));
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $entity = $this->entity;
    \assert($entity instanceof WorkspaceApprovalRequestInterface);
    $this->workflow = Workflow::load('workspace');
    \assert($this->workflow instanceof WorkflowInterface);
    $approvalRequestType = WorkspaceApprovalRequestType::load($entity->bundle());
    \assert($approvalRequestType instanceof WorkspaceApprovalRequestType);
    $this->transitions = \array_map(
      $this->workflow->getTypePlugin()->getTransition(...),
      $approvalRequestType->transitionIds(),
    );
    Inspector::assertAllObjects($this->transitions, EntityWorkflowTransition::class);

    $form['log'] = [
      '#type' => 'textarea',
      '#title' => $entity->get('revision_log_message')->getFieldDefinition()->getLabel(),
      '#description' => $entity->get('revision_log_message')->getFieldDefinition()->getDescription(),
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): ContentEntityInterface {
    foreach ($this->transitions as $transition) {
      \assert($transition instanceof EntityWorkflowTransition);
      if ($transition->hasLog() && $transition->isLogRequired() && empty($form_state->getValue('log'))) {
        $form_state->setError($form['log'], $this->t('Log field is required'));
      }
    }
    \assert($this->entity instanceof WorkspaceApprovalRequestInterface);
    return $this->entity;
  }

  /**
   * Checks access to the form.
   *
   * Suitable for a _custom_access route callback.
   * Based on the workspace_approval_request entity in the route.
   */
  public function access(AccountInterface $account, RouteMatchInterface $routeMatch): AccessResultInterface {
    $request = $routeMatch->getParameter('workspace_approval_request');
    if ($request instanceof WorkspaceApprovalRequestInterface) {
      return $request->access($this->operation, $account, TRUE);
    }
    return AccessResult::neutral();
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    parent::submitForm($form, $form_state);
    $entity = $this->getEntity();
    \assert($entity instanceof WorkspaceApprovalRequestInterface);
    $approvalType = $entity->getApprovalType();
    $workflowId = $approvalType->workflowId();
    \assert(\is_string($workflowId));
    $workspace = $entity->getSubject();
    $initialState = \entity_workflow_get_entity_state($workspace, $workflowId);
    $message = $form_state->getValue('log');
    $entity->set('revision_log_message', $message);
    $approvalStatus = $entity->getStatus();
    $this->entityPreSave($entity, $form, $form_state);
    $entity->validate();
    $entity->save();
    // Check whether the request was approved or rejected.
    if ($entity->getStatus() !== $approvalStatus && !$entity->isPending()) {
      $endState = \entity_workflow_get_entity_state($workspace, $workflowId);
      \assert(\is_string($initialState) && \is_string($endState));
      WorkflowTransitionLog::createFromEntity(
        $workspace,
        $workflowId,
        $initialState,
        $endState,
        'workspace_review',
        $message,
      )->save();
    }
    $this->entityPostSave($entity, $form, $form_state);
  }

  /**
   * Act on the request entity before it is validated and saved.
   */
  protected function entityPreSave(WorkspaceApprovalRequestInterface $entity, array &$form, FormStateInterface $form_state): void {
    // Override this method to update the request entity before it is validated
    // and saved.
  }

  /**
   * Act after the request entity has been validated and saved.
   */
  protected function entityPostSave(WorkspaceApprovalRequestInterface $entity, array &$form, FormStateInterface $form_state): void {
    // Override this method to do some work after the request entity has been
    // saved.
  }

}
