<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequest;
use Drupal\workspace_approval\Plugin\QueueWorker\ReviewerNotificationQueueWorker;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form controller for the workspace approval request entity edit forms.
 */
final class WorkspaceApprovalRequestForm extends ContentEntityForm {

  /**
   * {@inheritdoc}
   */
  public function __construct(
    EntityRepositoryInterface $entity_repository,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
    TimeInterface $time,
    private readonly QueueFactory $queue,
  ) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new self(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('queue'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state): int {
    \assert($this->entity instanceof WorkspaceApprovalRequestInterface);
    $result = parent::save($form, $form_state);

    $message_args = ['%label' => $this->entity->toLink()->toString()];
    $logger_args = [
      '%label' => $this->entity->label(),
      'link' => $this->entity->toLink($this->t('View'))->toString(),
    ];

    if ($result === \SAVED_NEW) {
      $this->messenger()->addStatus($this->t('New workspace approval request %label has been created.', $message_args));
      $this->logger('workspace_approval')->notice('New workspace approval request %label has been created.', $logger_args);
      if ($this->entity->isApprovalRequired()) {
        $data = ReviewerNotificationQueueWorker::createData($this->entity);
        $this->queue->get(ReviewerNotificationQueueWorker::ID)->createItem($data);
      }
    }
    elseif ($result === \SAVED_UPDATED) {
      $this->messenger()->addStatus($this->t('The workspace approval request %label has been updated.', $message_args));
      $this->logger('workspace_approval')->notice('The workspace approval request %label has been updated.', $logger_args);
    }
    else {
      throw new \LogicException('Could not save the entity.');
    }

    $form_state->setRedirect('entity.workspace_approval_request.list.workspace', [
      'workspace' => $this->entity->getSubject()->id(),
    ]);

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityFromRouteMatch(RouteMatchInterface $route_match, $entity_type_id) {
    $workspace = $route_match->getParameter('workspace');
    $entity = parent::getEntityFromRouteMatch($route_match, $entity_type_id);
    \assert($entity instanceof WorkspaceApprovalRequest);
    if ($entity->isNew()) {
      $entity->set('subject', $workspace);
    }
    return $entity;
  }

}
