<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\Core\Utility\Error;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;

/**
 * Provides a confirmation form to reject an approval request.
 */
final class WorkspaceApprovalRequestRejectForm extends WorkspaceApprovalRequestConfirmForm {

  /**
   * {@inheritdoc}
   */
  protected $operation = 'reject';

  /**
   * {@inheritdoc}
   */
  public function getQuestion(): TranslatableMarkup {
    return $this->t('Reject this %title?', [
      '%title' => $this->getBundleEntity()?->label(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription(): TranslatableMarkup {
    return $this->t("This action will prevent this content from transitioning to its next workflow state.");
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl(): Url {
    return Url::fromUserInput($this->getRedirectDestination()->get())->setAbsolute();
  }

  /**
   * {@inheritdoc}
   */
  protected function entityPreSave(WorkspaceApprovalRequestInterface $entity, array &$form, FormStateInterface $form_state): void {
    $entity->set('approved', FALSE);
  }

  /**
   * {@inheritdoc}
   */
  protected function entityPostSave(WorkspaceApprovalRequestInterface $entity, array &$form, FormStateInterface $form_state): void {
    $workspace = $entity->getSubject();
    $form_state->setRedirectUrl(Url::fromRoute(
      'entity.workspace_approval_request.list.workspace',
      ['workspace' => $workspace->id()],
    ));
    try {
      \assert($this->workflow instanceof WorkflowInterface);
      $log = (string) $this->t('Automatically reverted due to a rejected approval request.');
      \entity_workflow_transition($workspace, $this->workflow->id(), 'back_to_draft', $log);
      $this->messenger()->addMessage($this->t('The workflow state has been changed.'));
    }
    catch (\Exception $e) {
      Error::logException($this->logger('entity_workflow'), $e);
      $this->messenger()->addError($this->t('An error occurred while changing the workflow state. See logs for more details.'));
    }
  }

}
