<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Form;

use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Config\Entity\ConfigEntityStorageInterface;
use Drupal\Core\Entity\BundleEntityFormBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityReferenceSelection\SelectionInterface;
use Drupal\Core\Entity\EntityReferenceSelection\SelectionPluginManagerInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Drupal\workspace_approval\WorkspaceApprovalNotifications;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form handler for workspace approval request type forms.
 */
final class WorkspaceApprovalRequestTypeForm extends BundleEntityFormBase {

  /**
   * Constructs a new WorkspaceApprovalRequestTypeForm object.
   */
  public function __construct(
    protected SelectionPluginManagerInterface $selectionPluginManager,
    EntityTypeManagerInterface $entityTypeManager,
  ) {
    // Cannot use constructor promotion since this property is defined by a
    // parent class.
    $this->entityTypeManager = $entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new self(
      $container->get('plugin.manager.entity_reference_selection'),
      $container->get('entity_type.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state): array {
    \assert($this->entity instanceof WorkspaceApprovalRequestType);
    $form = parent::form($form, $form_state);

    if ($this->operation === 'edit') {
      $form['#title'] = $this->t('Edit %label workspace approval request type', ['%label' => $this->entity->label()]);
    }

    $form['label'] = [
      '#title' => $this->t('Label'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->label(),
      '#description' => $this->t('The human-readable name of this workspace approval request type.'),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#maxlength' => EntityTypeInterface::BUNDLE_MAX_LENGTH,
      '#machine_name' => [
        'exists' => [WorkspaceApprovalRequestType::class, 'load'],
        'source' => ['label'],
      ],
      '#description' => $this->t('A unique machine-readable name for this workspace approval request type. It must only contain lowercase letters, numbers, and underscores.'),
    ];

    $workflows = $this->getWorkflowOptions();
    $workflow_id = $form_state->getValue('workflow')
      ?? $this->entity->workflowId()
      ?? \array_key_first($workflows);

    $form['workflow'] = [
      '#title' => $this->t('Workflow'),
      '#type' => 'select',
      '#default_value' => $workflow_id,
      '#options' => $workflows,
      '#description' => $this->t('Select the workflow which requires these reviews.'),
      // Update the transition select options below when the selected workflow
      // changes.
      '#ajax' => [
        'callback' => [$this, 'getWorkflowTransitionSelectAjax'],
        'event' => 'change',
        'wrapper' => 'workflow-transition-select',
      ],
    ];

    $form['transition_select'] = $this->getWorkflowTransitionSelect($form_state);

    $message_template_options = $this->getMessageTemplateOptions();

    $form['review_requested_template'] = [
      '#type' => 'select2',
      '#title' => $this->t('Reviewer request message'),
      '#description' => $this->t('The message template used to notify reviewers of a request to provide this review.'),
      '#cardinality' => 1,
      '#empty_option' => $this->t('- Do not send a notification message -'),
      '#default_value' => $this->entity->messageTemplate(WorkspaceApprovalNotifications::ReviewRequested),
      '#options' => $message_template_options,
    ];

    $form['review_provided_template'] = [
      '#type' => 'select2',
      '#title' => $this->t('New review message'),
      '#description' => $this->t('The message template used to notify requesters that a review has been provided.'),
      '#cardinality' => 1,
      '#empty_option' => $this->t('- Do not send a notification message -'),
      '#default_value' => $this->entity->messageTemplate(WorkspaceApprovalNotifications::ReviewProvided),
      '#options' => $message_template_options,
    ];

    return $this->protectBundleIdElement($form);
  }

  /**
   * Gets a select list options array of available message templates.
   *
   * @return array
   *   An array of message template labels keyed by message template ID.
   */
  private function getMessageTemplateOptions(): array {
    $template_storage = $this->entityTypeManager->getStorage('message_template');
    \assert($template_storage instanceof ConfigEntityStorageInterface);
    $options = [];
    foreach ($template_storage->loadMultiple() as $template) {
      $options[$template->id()] = $template->label();
    }
    return $options;
  }

  /**
   * Gets a select element to choose an applicable workflow transition.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   *
   * @return array
   *   A render array defining a select list element of available terms, limited
   *   to the chosen vocabulary.
   */
  private function getWorkflowTransitionSelect(FormStateInterface $form_state): array {
    \assert($this->entity instanceof WorkspaceApprovalRequestType);
    $elements = [
      '#type' => 'container',
      '#attributes' => ['id' => 'workflow-transition-select'],
    ];
    $workflows = $this->getWorkflowOptions();
    $workflow_id = $form_state->getValue('workflow')
      ?? $this->entity->workflowId()
      ?? \array_key_first($workflows);
    $options = $this->getWorkflowTransitionOptions($workflow_id);
    $elements['guarded_transitions'] = [
      '#title' => $this->t('Guarded transitions'),
      '#type' => 'checkboxes',
      '#default_value' => $this->entity->transitionIds(),
      '#description' => $this->t('Select the specific transition which requires these reviews. This transition will be disabled until the content has been approved by all assigned reviewers.'),
      '#options' => $options,
    ];
    return $elements;
  }

  /**
   * Gets an updated transition selection element.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   *
   * @return array
   *   A render array to set via AJAX.
   */
  public function getWorkflowTransitionSelectAjax(array $form, FormStateInterface $form_state): array {
    return $form['transition_select'];
  }

  /**
   * Gets a select list options array of workflow transitions.
   *
   * @param string $workflow_id
   *   The workflow to return options for.
   *
   * @return array
   *   An array of workflow transition labels keyed by transition ID.
   */
  private function getWorkflowTransitionOptions(string $workflow_id): array {
    $workflow_storage = $this->entityTypeManager->getStorage('workflow');
    \assert($workflow_storage instanceof ConfigEntityStorageInterface);
    $workflow = $workflow_storage->load($workflow_id);
    \assert($workflow instanceof WorkflowInterface);
    $workflow_type = $workflow->getTypePlugin();
    $transitions = $workflow_type->getTransitions();
    $options = [];
    foreach ($transitions as $id => $transition) {
      $options[$id] = $transition->label();
    }
    return $options;
  }

  /**
   * Gets a select list options array of workflow.
   *
   * @return \Drupal\workflows\WorkflowInterface[]
   *   An array of workflows keyed by workflow ID.
   */
  private function getWorkflowOptions(): array {
    $handlers = $this->selectionPluginManager->getSelectionGroups('workflow');
    $handler_config = $handlers['default']['default:workflow'] ?? NULL;
    if (\is_null($handler_config)) {
      throw new \LogicException('The workflow entity type does not have a default selection handler');
    }
    $handler = $this->selectionPluginManager->getInstance([
      'target_type' => 'workflow',
    ] + $handler_config);
    \assert($handler instanceof SelectionInterface);
    $entities_by_bundle = $handler->getReferenceableEntities();
    return $entities_by_bundle['workflow'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state): array {
    $actions = parent::actions($form, $form_state);
    $actions['submit']['#value'] = $this->t('Save workspace approval request type');
    $actions['delete']['#value'] = $this->t('Delete workspace approval request type');
    return $actions;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state): int {
    $result = parent::save($form, $form_state);
    $message_args = ['%label' => $this->entity->label()];
    $this->messenger()->addStatus(
      match($result) {
        \SAVED_NEW => $this->t('The workspace approval request type %label has been added.', $message_args),
        \SAVED_UPDATED => $this->t('The workspace approval request type %label has been updated.', $message_args),
        default => throw new \DomainException()
      }
    );
    $form_state->setRedirectUrl($this->entity->toUrl('collection'));

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  protected function copyFormValuesToEntity(EntityInterface $entity, array $form, FormStateInterface $form_state): void {
    foreach ([
      'id' => 'id',
      'label' => 'label',
      'workflow' => 'workflow',
      'guarded_transitions' => 'guarded_transitions',
    ] as $property => $path) {
      \assert($entity instanceof ConfigEntityInterface);
      $value = $form_state->getValue($path);
      if ($path === 'guarded_transitions') {
        $value = \array_filter($value);
      }
      $entity->set($property, $value);
    }
    $message_templates = [];
    foreach (\array_column(WorkspaceApprovalNotifications::cases(), 'value') as $message_type) {
      $message_templates[$message_type] = $form_state->getValue("{$message_type}_template", '');
    }
    $entity->set('message_templates', $message_templates);
  }

}
