<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Plugin\EntityReferenceSelection;

use Drupal\Core\Entity\Attribute\EntityReferenceSelection;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\user\Plugin\EntityReferenceSelection\UserSelection;
use Drupal\user\RoleInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequest;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;
use Drupal\workspaces\WorkspaceInterface;

/**
 * Provides specific access control for reviewer selection.
 */
#[EntityReferenceSelection(
  id: "workspace_approval:reviewer",
  label: new TranslatableMarkup("Reviewer selection"),
  group: "workspace_approval",
  weight: 0,
  entity_types: ["user"],
)]
final class ReviewerSelection extends UserSelection {

  /**
   * Config key indicating whether to include users with active requests.
   */
  public const string INCLUDE_ACTIVE_REVIEWERS_KEY = 'include_active_reviewers';

  /**
   * {@inheritdoc}
   */
  public function getConfiguration(): array {
    $entity = $this->configuration['entity'] ?? NULL;
    \assert(\is_null($entity) || $entity instanceof WorkspaceApprovalRequest);
    return [
      'target_type' => 'user',
    ] + $this->configuration + [
      self::INCLUDE_ACTIVE_REVIEWERS_KEY => FALSE,
      'request_type' => $entity?->bundle(),
      'subject' => $entity?->getSubject(),
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function buildEntityQuery($match = NULL, $match_operator = 'CONTAINS'): QueryInterface {
    $query = parent::buildEntityQuery($match, $match_operator);
    $reviewerRoles = $this->getReviewerRoles();
    if (empty($reviewerRoles)) {
      // If there are no roles that can approve requests of the given type, then
      // there can't be any users with that permission, so return an empty
      // array.
      return $query;
    }
    $query = $query
      ->condition('status', 1)
      ->condition('roles', $reviewerRoles, 'IN');
    $configuration = $this->getConfiguration();
    $excludeActiveReviewers = !($configuration[self::INCLUDE_ACTIVE_REVIEWERS_KEY] ?? FALSE);
    if ($excludeActiveReviewers) {
      $existingReviewerIds = $this->getExistingReviewerIds();
      if (!empty($existingReviewerIds)) {
        $userEntityType = $this->entityTypeManager->getDefinition('user');
        \assert($userEntityType instanceof EntityTypeInterface);
        $query = $query->condition((string) $userEntityType->getKey('id'), $existingReviewerIds, 'NOT IN');
      }
    }
    return $query;
  }

  /**
   * Gets all user IDs that have a pending request for the configured subject.
   *
   * @return array
   *   An array of user IDs identifying users that have a pending request for
   *   the configured review type and subject.
   */
  private function getExistingReviewerIds(): array {
    $reviewSubject = $this->getConfiguration()['subject'];
    if (!$reviewSubject) {
      return [];
    }
    \assert($reviewSubject instanceof WorkspaceInterface);
    $requestStorage = $this->entityTypeManager->getStorage('workspace_approval_request');
    $entityType = $this->entityTypeManager->getDefinition('workspace_approval_request');
    \assert($entityType instanceof EntityTypeInterface);
    $requestType = $this->getConfiguration()['request_type'];
    $statusKey = $entityType->getKey('published');
    $bundleKey = $entityType->getKey('bundle');
    \assert(\is_string($statusKey) && \is_string($bundleKey));
    $pending = $requestStorage->getQuery()
      ->condition('subject', $reviewSubject->id())
      ->condition($statusKey, TRUE)
      ->condition($bundleKey, $requestType)
      ->accessCheck(FALSE)
      ->execute();
    $reviewerIds = [];
    foreach ($requestStorage->loadMultiple($pending) as $request) {
      \assert($request instanceof WorkspaceApprovalRequestInterface);
      $reviewerIds[] = (int) $request->getReviewer()->id();
    }
    return $reviewerIds;
  }

  /**
   * Gets roles with permission to approve the configured review request type.
   */
  private function getReviewerRoles(): array {
    $requestType = $this->entityTypeManager
      ->getStorage('workspace_approval_request_type')
      ->load($this->getConfiguration()['request_type']);
    \assert($requestType instanceof WorkspaceApprovalRequestType);
    $permission_id = "approve {$requestType->id()} workspace approval requests";
    $roles = $this->entityTypeManager->getStorage('user_role')->loadMultiple();
    $reviewerRoles = \array_filter(
      $roles,
      static fn(RoleInterface $role) => $role->hasPermission($permission_id) && !$role->isAdmin(),
    );
    return \array_keys($reviewerRoles);
  }

}
