<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Plugin\QueueWorker;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Queue\QueueWorkerBase;
use Drupal\message\Entity\Message;
use Drupal\message_notify\MessageNotifyInterface;
use Drupal\workspace_approval\WorkspaceApprovalNotifications;
use Drupal\workspace_approval\WorkspaceApprovalRequestField;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Process a queue of workspace transitions to send notifications.
 */
abstract class NotificationQueueWorkerBase extends QueueWorkerBase implements ContainerFactoryPluginInterface {

  /**
   * The notification type sent by this worker.
   */
  protected ?WorkspaceApprovalNotifications $notificationType = NULL;

  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected LoggerChannelFactoryInterface $loggerFactory,
    protected MessageNotifyInterface $messageNotifier,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    if (\is_null($this->notificationType)) {
      throw new \LogicException('$notificationType must be set.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('logger.factory'),
      $container->get('message_notify.sender'),
    );
  }

  /**
   * Creates a queue item data array.
   *
   * @param \Drupal\workspace_approval\WorkspaceApprovalRequestInterface $request
   *   The request related to the notification.
   *
   * @return array
   *   An array suitable for QueueInterface::createItem() for this queue.
   */
  public static function createData(WorkspaceApprovalRequestInterface $request): array {
    return [
      'workspace_approval_request_id' => $request->id(),
    ];
  }

  /**
   * Loads the workspace approval request entity identified by the data.
   *
   * @param array $data
   *   The data array.
   *
   * @return \Drupal\workspace_approval\WorkspaceApprovalRequestInterface|null
   *   An entity object or NULL if one could not be found.
   */
  private function getRequestFromData(array $data): ?WorkspaceApprovalRequestInterface {
    $requestId = $data['workspace_approval_request_id'] ?? NULL;
    if (empty($requestId)) {
      $this->loggerFactory->get('workspace_approval')
        ->error(\sprintf('Incomplete data found on %s queue item', $this->getPluginId()));
      return NULL;
    }
    $request = $this->entityTypeManager->getStorage('workspace_approval_request')->load($requestId);
    if (!$request instanceof WorkspaceApprovalRequestInterface) {
      $this->loggerFactory->get('workspace_approval')
        ->error('Workspace approval request (%id) not found.', ['%id' => $data['workspace_approval_request_id']]);
      return NULL;
    }
    return $request;
  }

  /**
   * {@inheritdoc}
   */
  public function processItem($data): void {
    $workspace_approval_request = $this->getRequestFromData($data);
    if (empty($workspace_approval_request)) {
      return;
    }
    if (!$workspace_approval_request->isPublished()) {
      // Do not send any notifications for inactive reviews.
      return;
    }
    \assert($this->notificationType instanceof WorkspaceApprovalNotifications);
    $message_template = $workspace_approval_request->getMessageTemplate($this->notificationType);
    if (empty($message_template)) {
      return;
    }
    $uid_to_notify = $this->getUidToNotify($workspace_approval_request);
    $message = Message::create([
      'template' => $message_template,
      'uid' => $uid_to_notify,
    ]);
    $message->set(WorkspaceApprovalRequestField::FIELD_NAME, $workspace_approval_request);
    $message->save();
    $email_options = [];
    $this->messageNotifier->send($message, $email_options);
    $this->loggerFactory->get('workspace_approval')
      ->info(\sprintf('Email notification sent to user with uid %d.', $uid_to_notify));
  }

  /**
   * Gets the user id to notify.
   *
   * @param \Drupal\workspace_approval\WorkspaceApprovalRequestInterface $workspace_approval_request
   *   The workspace approval to examine.
   *
   * @return int
   *   The user id.
   */
  abstract protected function getUidToNotify(WorkspaceApprovalRequestInterface $workspace_approval_request): ?int;

}
