<?php

declare(strict_types=1);

namespace Drupal\workspace_approval;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\user\UserInterface;
use Drupal\workspaces\Entity\Workspace;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Helper class for Drupal hook implementations.
 */
final class WorkspaceApprovalEntityHooks implements ContainerInjectionInterface {

  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('entity_type.manager'),
    );
  }

  /**
   * Removes relevant workspace_approval_request entities.
   *
   * Namely, if a user is removed, remove any workspace approval request
   * entities that has that user as the reviewer.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The user entity.
   */
  public function onUserEntityDelete(EntityInterface $entity): void {
    if (!$entity instanceof UserInterface) {
      // Invalid case, nothing to do.
      return;
    }
    $storage = $this->entityTypeManager
      ->getStorage('workspace_approval_request');
    $result = $storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('reviewer', $entity->id())
      ->execute();
    $approvals = $storage->loadMultiple($result);
    foreach ($approvals as $approval) {
      $approval->delete();
    }
  }

  /**
   * Removes relevant workspace_approval_request entities.
   *
   * Namely, if a workspace is removed, remove any workspace approval request
   * entities associated with it.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The workspace entity.
   */
  public function onWorkspaceEntityDelete(EntityInterface $entity): void {
    if (!$entity instanceof Workspace) {
      // Invalid case, nothing to do.
      return;
    }
    $storage = $this->entityTypeManager
      ->getStorage('workspace_approval_request');
    $result = $storage->getQuery()
      ->accessCheck(FALSE)
      ->condition('subject', $entity->id())
      ->execute();
    $approvals = $storage->loadMultiple($result);
    foreach ($approvals as $approval) {
      $approval->delete();
    }
  }

}
