<?php

declare(strict_types=1);

namespace Drupal\workspace_approval;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\BundlePermissionHandlerTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides workspace approval permissions.
 */
final class WorkspaceApprovalPermissions implements ContainerInjectionInterface {

  use BundlePermissionHandlerTrait;
  use StringTranslationTrait;

  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
    TranslationInterface $stringTranslation,
  ) {
    $this->stringTranslation = $stringTranslation;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new self(
      $container->get('entity_type.manager'),
      $container->get('string_translation')
    );
  }

  /**
   * Gets all available permissions.
   *
   * @return array
   *   Permissions array. See PermissionHandlerInterface::getPermissions() for
   *   the array structure.
   */
  public function permissions(): array {
    return $this->generatePermissions(
      WorkspaceApprovalRequestType::loadMultiple(),
      $this->buildPermissions(...),
    );
  }

  /**
   * Gets all available permissions.
   *
   * @return array
   *   Permissions array. See PermissionHandlerInterface::getPermissions() for
   *   the array structure.
   */
  private function buildPermissions(WorkspaceApprovalRequestType $requestType): array {
    $bundle_of_entity_type_id = $requestType->getEntityType()->getBundleOf();
    \assert(\is_string($bundle_of_entity_type_id));
    $bundle_of = $this->entityTypeManager->getDefinition($bundle_of_entity_type_id);
    \assert($bundle_of instanceof EntityTypeInterface);
    $transitions = $this->getGuardedTransitions($requestType);
    $permissions = [];
    foreach ($transitions as $transition) {
      $permissions["approve {$requestType->id()} workspace approval requests"] = [
        'title' => $this->t('Approve %type @requests', [
          '%type' => $requestType->label(),
          '@requests' => $bundle_of->getPluralLabel(),
        ]),
        'description' => $this->t('Allow the user to review content prior to the %transition workflow transition for %type purposes.', [
          '%type' => $requestType->label(),
          '%transition' => $transition->label(),
        ]),
      ];
    }
    $permissions["create {$requestType->id()} workspace approval requests"] = [
      'title' => $this->t('Create %type @requests', [
        '%type' => $requestType->label(),
        '@requests' => $bundle_of->getPluralLabel(),
      ]),
      'description' => $this->t('Allow the user to create %type approval requests.', [
        '%type' => $requestType->label(),
      ]),
    ];
    return $permissions;
  }

  /**
   * Gets the workflow transition guarded by the given request type.
   *
   * @param \Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType $requestType
   *   The request type that guards a transition.
   *
   * @return \Drupal\workflows\TransitionInterface[]
   *   The guarded transition objects.
   */
  private function getGuardedTransitions(WorkspaceApprovalRequestType $requestType): array {
    $storage = $this->entityTypeManager->getStorage('workflow');
    $workflow = $storage->load($requestType->workflowId());
    $transitions = $requestType->transitionIds();
    \assert($workflow instanceof WorkflowInterface);
    return \array_map(
      $workflow->getTypePlugin()->getTransition(...),
      $transitions,
    );
  }

}
