<?php

declare(strict_types=1);

namespace Drupal\workspace_approval;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Entity\EntityAccessControlHandler;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Session\AccountInterface;

/**
 * Controls access to workspace approval request entities.
 */
final class WorkspaceApprovalRequestAccessControlHandler extends EntityAccessControlHandler {

  /**
   * {@inheritdoc}
   */
  protected function checkAccess(EntityInterface $entity, $operation, AccountInterface $account) {
    \assert($entity instanceof WorkspaceApprovalRequestInterface);
    $default = parent::checkAccess($entity, $operation, $account);
    $account_is_requester = $account->id() === $entity->getOwnerId();
    $account_is_reviewer = $account->id() === $entity->getReviewer()->id();
    $operationResult = match ($operation) {
      // The "dismiss" operation is the same as "unpublish" so check the
      // "update" operation with the parent. And only the original requester
      // should be allowed to dismiss a review.
      'dismiss' => parent::checkAccess($entity, 'update', $account)
        ->andIf(AccessResult::forbiddenIf(!$account_is_requester))
        // There's no reason to dismiss an approved review.
        ->andIf(AccessResult::forbiddenIf($entity->isApproved()))
        ->addCacheableDependency($entity),
      // Only reviewers should be able to approve reviews. And approved reviews
      // cannot be rejected twice.
      'approve' => AccessResult::forbiddenIf(!$account_is_reviewer)
        ->andIf(AccessResult::forbiddenIf($entity->isApproved()))
        ->addCacheableDependency($entity),
      // Only reviewers should be able to reject reviews. And rejected reviews
      // cannot be rejected twice.
      'reject' => AccessResult::forbiddenIf(!$account_is_reviewer)
        ->andIf(AccessResult::forbiddenIf($entity->isRejected()))
        ->addCacheableDependency($entity),
      // Only the original requester should be able to delete their request.
      'delete' => AccessResult::forbiddenIf(!$account_is_requester)
        ->addCacheableDependency($entity),
      // Unrecognized operation. This allows the orIf() below to fall back to
      // the default result.
      default => AccessResult::neutral(),
    };
    // This AccessResult->orIf()->orIf() chain will allow access unless the
    // operation-specific or default result is explicitly forbidden.
    return AccessResult::allowed()->orIf($operationResult)->orIf($default);
  }

  /**
   * {@inheritdoc}
   */
  protected function checkCreateAccess(AccountInterface $account, array $context, $entity_bundle = NULL): AccessResultInterface {
    return parent::checkCreateAccess($account, $context, $entity_bundle)
      ->orIf(AccessResult::allowedIfHasPermissions($account, [
        'create ' . $entity_bundle . ' workspace approval requests',
      ]));
  }

}
