<?php

declare(strict_types=1);

namespace Drupal\workspace_approval;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityChangedInterface;
use Drupal\Core\Entity\EntityPublishedInterface;
use Drupal\user\EntityOwnerInterface;
use Drupal\user\UserInterface;
use Drupal\workflows\TransitionInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Drupal\workspaces\WorkspaceInterface;

/**
 * Provides an interface defining a workspace approval request entity type.
 */
interface WorkspaceApprovalRequestInterface extends ContentEntityInterface, EntityOwnerInterface, EntityPublishedInterface, EntityChangedInterface {

  /**
   * Get the user assigned to perform this review.
   */
  public function getReviewer(): UserInterface;

  /**
   * Get the workspace under review.
   */
  // phpcs:ignore
  public function getSubject(): WorkspaceInterface;

  /**
   * Get request status.
   */
  public function getStatus(): ?bool;

  /**
   * Determine if request is approved.
   */
  public function isApproved(): bool;

  /**
   * Determine if request is rejected.
   */
  public function isRejected(): bool;

  /**
   * Determine if request is pending.
   */
  public function isPending(): bool;

  /**
   * Gets the template to use for the given notification type for this request.
   *
   * @return ?string
   *   The message template ID or NULL if no notification should be sent.
   */
  public function getMessageTemplate(WorkspaceApprovalNotifications $type): ?string;

  /**
   * Gets the bundle entity.
   *
   * @return \Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType
   *   The bundle entity representing the workspace approval request type.
   */
  public function getApprovalType(): WorkspaceApprovalRequestType;

  /**
   * Determine if this request must be approved for a transition.
   *
   * @param ?\Drupal\workflows\WorkflowInterface $workflow
   *   (optional) The workflow to consider.
   * @param ?\Drupal\workflows\TransitionInterface $transition
   *   (optional) The only transition to consider.
   *
   * @return bool
   *   TRUE if this request must be approved before the subject workspace may
   *   be transitioned; FALSE otherwise. If no transition is specified, will be
   *   evaluated for all of the subject's available transitions. Note: always
   *   returns FALSE if the request has already been approved.
   */
  public function isApprovalRequired(?WorkflowInterface $workflow = NULL, ?TransitionInterface $transition = NULL): bool;

}
