<?php

declare(strict_types=1);

namespace Drupal\workspace_approval;

use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityListBuilder;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class a workspace approval request list builder.
 */
abstract class WorkspaceApprovalRequestListBuilderBase extends EntityListBuilder implements ContainerInjectionInterface {

  public function __construct(
    EntityTypeInterface $entity_type,
    EntityStorageInterface $storage,
    protected readonly RouteMatchInterface $routeMatch,
    private readonly AccountProxyInterface $currentUser,
  ) {
    parent::__construct($entity_type, $storage);
  }

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity_type.manager')->getStorage($entity_type->id()),
      $container->get('current_route_match'),
      $container->get('current_user')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return self::createInstance(
      $container,
      $container->get('entity_type.manager')->getDefinition('workspace_approval_request'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getTitle(): TranslatableMarkup|null|string {
    $entity = $this->routeMatch->getParameter('workspace');
    \assert($entity instanceof EntityInterface);
    return $entity->label();
  }

  /**
   * {@inheritdoc}
   */
  public function buildHeader(): array {
    $header = [];
    $header['request'] = $this->t('Request');
    $header['uid'] = $this->t('Requester');
    $header['reviewer'] = $this->t('Approver');
    $header['created'] = $this->t('Requested');
    $header['status'] = $this->t('Status');
    $header['operations'] = $this->t('Operations');
    return $header + parent::buildHeader();
  }

  /**
   * {@inheritdoc}
   */
  public function buildRow(EntityInterface $entity): array {
    $row = [];
    /** @var \Drupal\workspace_approval\WorkspaceApprovalRequestInterface $entity */
    $row['request'] = $entity->toLink();
    $username_options = [
      'label' => 'hidden',
      'settings' => ['link' => $entity->get('uid')->entity->isAuthenticated()],
    ];
    $row['uid']['data'] = $entity->get('uid')->view($username_options);
    $row['reviewer']['data'] = $entity->get('reviewer')->view($username_options);
    $row['created']['data'] = $entity->get('created')->view(['label' => 'hidden']);
    $approved = $entity->get('approved');
    \assert($approved instanceof FieldItemListInterface);

    $row['status']['data'] = $this->t('%status', [
      '%status' => match($entity->getStatus()) {
        NULL => $this->t('Pending'),
        FALSE => $approved->getSetting('off_label'),
        TRUE => $approved->getSetting('on_label'),
      },
    ]);
    $row['operations']['data'] = $this->buildOperations($entity);
    return $row + parent::buildRow($entity);
  }

  /**
   * {@inheritdoc}
   */
  protected function getEntityListQuery(): QueryInterface {
    return parent::getEntityListQuery()
      ->condition('active', TRUE)
      ->sort('created', 'DESC');
  }

  /**
   * {@inheritdoc}
   */
  protected function getDefaultOperations(EntityInterface $entity): array {
    \assert($entity instanceof WorkspaceApprovalRequestInterface);
    $operations = parent::getDefaultOperations($entity);
    $additionalOps = [];
    foreach (['approve', 'reject', 'dismiss'] as $op) {
      $has_form = $entity->hasLinkTemplate("{$op}-form");
      if ($has_form) {
        $allowed = Url::fromRoute("entity.workspace_approval_request.{$op}_form", [
          'workspace_approval_request' => $entity->id(),
        ])->access($this->currentUser, TRUE);
        \assert($allowed instanceof AccessResultInterface);
        if ($allowed->isAllowed()) {
          $link_title = match ($op) {
            'approve' => $this->t('Approve'),
            'reject' => $this->t('Reject'),
            'dismiss' => $this->t('Dismiss'),
          };
          $additionalOps[$op] = [
            'title' => $link_title,
            'url' => $entity->toUrl("{$op}-form"),
          ];
        }
      }
    }
    return $additionalOps + $operations;
  }

}
