<?php

declare(strict_types=1);

namespace Drupal\workspace_approval;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Url;
use Drupal\entity_workflow\StateTransitionValidatorInterface;
use Drupal\workflows\TransitionInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspaces\WorkspaceInterface;

/**
 * Validates workspace workflow transitions based on their review status.
 */
final class WorkspaceApprovalStateTransitionValidator implements StateTransitionValidatorInterface {

  use StringTranslationTrait;

  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly MessengerInterface $messenger,
    TranslationInterface $stringTranslation,
    private readonly StateTransitionValidatorInterface $innerValidator,
  ) {
    $this->stringTranslation = $stringTranslation;
  }

  /**
   * {@inheritdoc}
   */
  public function getValidTransitions(ContentEntityInterface $entity, WorkflowInterface $workflow, AccountInterface $user, ?WorkspaceInterface $workspace = NULL): array {
    $transitions = $this->innerValidator->getValidTransitions($entity, $workflow, $user, $workspace);
    $valid = [];
    $invalid = [];
    foreach ($transitions as $transition) {
      if ($this->hasUnapprovedRequests($entity, $workflow, $transition)) {
        $invalid[$transition->id()] = $transition;
      }
      else {
        $valid[$transition->id()] = $transition;
      }
    }
    // If $workspace is NULL, then this method was probably called in the
    // context of EntityWorkflowWorkspaceListBuilder::offCanvasRender. In that
    // case, it does not make sense to add a warning message because the message
    // will not appear until the next page load.
    if (!empty($invalid) && !\is_null($workspace)) {
      $get_label = static fn (TransitionInterface $t) => $t->label();
      $blocked = \array_map($get_label, $invalid);
      $message = $this->t(
        'The following workflow transitions are blocked because they have <a href=":request_list">a pending or rejected approval request</a>: %blocked',
        [
          '%blocked' => \implode(', ', $blocked),
          ':request_list' => Url::fromRoute('entity.workspace_approval_request.list.workspace', [
            'workspace' => $entity->id(),
          ])->toString(),
        ],
      );
      $this->messenger->addWarning($message);
    }
    return $valid;
  }

  /**
   * Whether the given subject has unapproved requests.
   */
  private function hasUnapprovedRequests(ContentEntityInterface $subject, WorkflowInterface $workflow, TransitionInterface $transition): bool {
    $applicable_types = $this->getApplicableRequestTypes($workflow, $transition);
    $is_guarded = !empty($applicable_types);
    if (!$is_guarded) {
      return FALSE;
    }
    $storage = $this->entityTypeManager->getStorage('workspace_approval_request');
    \assert($storage instanceof EntityStorageInterface);
    $query = $storage->getQuery();
    $unapproved = $query
      ->condition('bundle', $applicable_types, 'IN')
      ->condition('active', TRUE)
      ->condition('subject.target_id', $subject->id())
      ->condition(
        $query->orConditionGroup()
          ->condition('approved', NULL, 'IS NULL')
          ->condition('approved', FALSE)
      )
      ->accessCheck(FALSE)
      ->count()
      ->execute();
    return $unapproved > 0;
  }

  /**
   * Gets the request types (bundles) applicable to a workflow transition.
   */
  private function getApplicableRequestTypes(WorkflowInterface $workflow, TransitionInterface $transition): array {
    $storage = $this->entityTypeManager->getStorage('workspace_approval_request_type');
    \assert($storage instanceof EntityStorageInterface);
    $query = $storage
      ->getQuery()
      ->condition('workflow', $workflow->id())
      ->accessCheck(FALSE);
    // The OR condition group is necessary here, otherwise *every* guarded
    // transition would need to match $transition, which would be illogical.
    $matchAnyGuardedTransition = $query->orConditionGroup()->condition('guarded_transitions.*', $transition->id());
    $query->condition($matchAnyGuardedTransition);
    return $query->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function isTransitionValid(ContentEntityInterface $entity, WorkflowInterface $workflow, $transition_id, AccountInterface $user, ?WorkspaceInterface $workspace = NULL): bool {
    $valid = $this->innerValidator->isTransitionValid($entity, $workflow, $transition_id, $user, $workspace);
    if (!$valid) {
      return $valid;
    }
    $transition = $workflow->getTypePlugin()->getTransition($transition_id);
    $blocked = $this->hasUnapprovedRequests($entity, $workflow, $transition);
    if ($blocked) {
      $message = $this->t(
        'The %blocked transition is blocked for the %entity %type because it has <a href=":request_list">a pending or rejected approval request</a>.',
        [
          '%blocked' => $transition->label(),
          '%entity' => $entity->label(),
          '%type' => $entity->getEntityType()->getSingularLabel(),
          ':request_list' => Url::fromRoute('entity.workspace_approval_request.list.workspace', [
            'workspace' => $entity->id(),
          ])->toString(),
        ],
      );
      $this->messenger->addWarning($message);
    }
    return !$blocked;
  }

}
