<?php

declare(strict_types=1);

namespace Drupal\workspace_approval\Functional;

use Drupal\Component\Utility\Random;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\node\Traits\ContentTypeCreationTrait;
use Drupal\Tests\wse\Functional\WseTestUtilities;
use Drupal\node\NodeTypeInterface;
use Drupal\user\UserInterface;
use Drupal\workspace_approval\Entity\WorkspaceApprovalRequestType;
use Drupal\workspace_approval\WorkspaceApprovalNotifications;

/**
 * Tests the user workflow for requesting and performing workspace reviews.
 *
 * @group workspace_approval
 */
final class WorkspaceApprovalTest extends BrowserTestBase {

  use ContentTypeCreationTrait;
  use WseTestUtilities;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'block',
    'entity',
    'workspace_approval',
    'node',
    'toolbar',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * A node type for test content.
   */
  private NodeTypeInterface $nodeType;

  /**
   * A test review type.
   */
  private WorkspaceApprovalRequestType $reviewType;

  /**
   * The user who will modify a workspace and request a review.
   */
  private UserInterface $editor;

  /**
   * The user who will be requested to review a workspace.
   */
  private UserInterface $reviewerA;

  /**
   * Another user who can review workspaces but who will not have any requests.
   */
  private UserInterface $reviewerB;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    // Set up a custom review type.
    $this->reviewType = WorkspaceApprovalRequestType::create([
      'id' => 'standard_review',
      'label' => 'Standard review',
      'workflow' => 'workspace',
      'guarded_transitions' => ['approve'],
      'message_templates' => [
        WorkspaceApprovalNotifications::ReviewRequested->value => '',
        WorkspaceApprovalNotifications::ReviewProvided->value => '',
      ],
    ]);
    $this->reviewType->save();
    // Set up a node type to be modified in the test workspace.
    $this->nodeType = $this->drupalCreateContentType([
      'type' => $this->randomMachineName(),
      'label' => $this->getRandomGenerator()->word(8),
    ]);
    $permissions = [
      'access administration pages',
      'administer site configuration',
      'create workspace',
      'edit own workspace',
      'edit any workspace',
      'view own workspace',
      'administer workspaces',
      'access toolbar',
      'bypass entity access own workspace',
      'use all workspace transitions',
      "approve {$this->reviewType->id()} workspace approval requests",
    ];
    // Set up test users.
    $this->editor = $this->createUser($permissions);
    $this->reviewerA = $this->createUser($permissions);
    $this->reviewerB = $this->createUser($permissions);
    // Place these admin blocks to help navigate the functional test cases.
    $this->drupalPlaceBlock('local_actions_block');
    $this->drupalPlaceBlock('local_tasks_block');
  }

  /**
   * Tests that the test user can view a list of their own approval requests.
   */
  public function testMyReviewsPage(): void {
    // Log in as a content editor.
    $this->drupalLogin($this->editor);
    // Create and activate a new workspace.
    $this->wseCreateAndActivateWorkspaceThroughUi('Workspace 1', 'workspace_1');
    // Make a content change.
    $this->createNodeThroughUi($this->randomString(), (string) $this->nodeType->id());
    // Visit the workspace admin UI.
    $this->drupalGet('/admin/config/workflow/workspaces');
    // Navigate to the workspace workflow page.
    $this->clickLink('Workspace 1', 1);
    $this->clickLink('Workspace workflow');
    // Submit the workspace for review and assign a review to reviewerA.
    $this->submitForm([
      'transition' => 'submit_for_review',
      'standard_review[' . $this->reviewerA->id() . ']' => $this->reviewerA->id(),
    ], 'Change');
    // The next transition should be blocked because the workspace now has a
    // pending review.
    $this->assertSession()->pageTextContains('The following workflow transitions are blocked because they have a pending or rejected approval request: Approve');
    // Visit the workspace approvals to verify the request was created.
    $this->clickLink('Workspace approvals');
    $this->assertSession()->pageTextContains("Workspace 1 workspace standard review");
    $this->assertSession()->pageTextContains("Pending");
    // Verify that the "Approve" transition action is missing too.
    $this->clickLink('Manage');
    $this->assertSession()->linkNotExists('Approve');
    // Log out so the reviewer can log in.
    $this->drupalLogout();

    // Log in as the reviewer who was *not* requested to review the workspace.
    $this->drupalLogin($this->reviewerB);
    // Visit the workspace admin UI.
    $this->drupalGet('/admin/config/workflow/workspaces');
    // Navigate to the "My reviews" page.
    $this->clickLink("My reviews");
    // Verify that the user does not have any pending requests.
    $this->assertSession()->pageTextContains('You do not have any pending review requests at this time.');
    $this->drupalLogout();

    // Log in as the reviewer who *was* requested to review the workspace.
    $this->drupalLogin($this->reviewerA);
    // Visit the workspace admin UI.
    $this->drupalGet('/admin/config/workflow/workspaces');
    // Navigate to the "My reviews" page.
    $this->clickLink("My reviews");
    // Verify that the user has a pending request.
    $this->assertSession()->pageTextContains("Workspace 1 workspace standard review");
    $this->assertSession()->pageTextContains("Pending");
    // Approve the request.
    $this->clickLink("Approve");
    $remarks = (new Random())->sentences(24);
    $this->submitForm([
      'edit-log' => $remarks,
    ], "Confirm");
    $this->drupalLogout();

    // Log in as a content editor again.
    $this->drupalLogin($this->editor);
    // Visit the workspace admin UI.
    $this->drupalGet('/admin/config/workflow/workspaces');
    // Navigate to the test workspace.
    $this->clickLink('Workspace 1');
    // The next transition should be unblocked because the workspace no longer
    // has any pending or rejected review.
    $this->assertSession()->linkExists('Approve');
    // Navigate to the test workspace's workflow page.
    $this->clickLink('Workspace workflow');
    // Verify the review remarks are displayed.
    $this->assertSession()->pageTextContains($remarks);
    // Submit the workspace for approval and assign a review to reviewerA.
    $this->submitForm([
      'transition' => 'approve',
    ], 'Change');
    // Submit the workspace for publication.
    $this->submitForm([
      'transition' => 'publish',
    ], 'Change');
    // Verify the workspace was published.
    $this->assertSession()->pageTextContains('The workflow state has been changed.');
    $this->assertSession()->pageTextContains('Published');
  }

}
