<?php

/**
 * @file
 * Builds placeholder replacement tokens for workspace approval data.
 */

declare(strict_types=1);

use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\message\MessageInterface;
use Drupal\workspace_approval\WorkspaceApprovalRequestField;
use Drupal\workspace_approval\WorkspaceApprovalRequestInterface;

/**
 * Implements hook_token_info().
 */
function workspace_approval_token_info(): array {
  $token_type = \Drupal::service('token.entity_mapper')->getTokenTypeForEntityType('workspace_approval_request');
  return [
    'tokens' => [
      'workspace_approval_request' => [
        'approval-type' => [
          'name' => new TranslatableMarkup('Approval type label'),
        ],
        'approval-status' => [
          'name' => new TranslatableMarkup('Approval status'),
        ],
      ],
      'message' => [
        'workspace_approval_request' => [
          'name' => new TranslatableMarkup('Workspace approval request'),
          'type' => $token_type,
          'nested' => TRUE,
        ],
      ],
    ],
  ];
}

/**
 * Implements hook_tokens().
 */
function workspace_approval_tokens(string $type, array $tokens, array $data, array $options, BubbleableMetadata $bubbleable_metadata): array {
  $replacements = [];
  $tokenService = \Drupal::token();
  if ($type === 'workspace_approval_request') {
    \assert($data['workspace_approval_request'] instanceof WorkspaceApprovalRequestInterface);
    foreach ($tokens as $name => $original) {
      if ($name === 'approval-status') {
        $replacements[$original] = match ($data['workspace_approval_request']->getStatus()) {
          NULL => new TranslatableMarkup('Pending'),
          FALSE => new TranslatableMarkup('Rejected'),
          TRUE => new TranslatableMarkup('Approved'),
        };
      }
      if ($name === 'approval-type') {
        $type = $data['workspace_approval_request']->getApprovalType();
        $bubbleable_metadata->addCacheableDependency($type);
        $replacements[$original] = $type->label();
      }
    }
  }
  if ($type === 'message') {
    \assert($data['message'] instanceof MessageInterface);
    if ($data['message']->hasField(WorkspaceApprovalRequestField::FIELD_NAME)) {
      $request_field = $data['message']->get(WorkspaceApprovalRequestField::FIELD_NAME);
      if (!$request_field->isEmpty()) {
        $request = $request_field->first()?->get('entity')->getValue();
        \assert($request instanceof WorkspaceApprovalRequestInterface);

        foreach ($tokens as $name => $original) {
          if ($name === 'workspace_approval_request') {
            $bubbleable_metadata->addCacheableDependency($request);
            $replacements[$original] = $request->label();
          }
        }

        $nestedTokens = $tokenService->findWithPrefix($tokens, 'workspace_approval_request');
        if (!empty($nestedTokens)) {
          $replacements += $tokenService->generate(
            'workspace_approval_request',
            $nestedTokens,
            ['workspace_approval_request' => $request],
            $options,
            $bubbleable_metadata,
          );
        }
      }
    }
  }
  return $replacements;
}
